import os
import time
import tkinter
import tkinter.filedialog
import uuid
from enum import Enum
from pathlib import Path

import numpy as np
from loguru import logger

from backend import load_map_from_ply_and_txt
from models.LogSeverity import LogSeverity
from models.LogRecord import LogRecord
from models.PointOfInterest import PointOfInterest
from models.Project import Project
import eel

from models.SegmentWall import SegmentWall
import base64
import sklearn.preprocessing

from models.StateADC import StateADC

path_current_dir = os.path.dirname(__file__)
path_base = f'{path_current_dir}/../'

class EnumDirectoryMode(Enum):
    MODE_PLY_TXT = 'MODE_PLY_TXT'
    MODE_PROJECT = 'MODE_PROJECT'

class ControllerChromium:
    __instance = None

    def __init__(self):
        if ControllerChromium.__instance is not None:
            raise Exception("Cannot have 2 instances of singleton ControllerChromium")

        self.project = Project()
        self.filepath_txt = ''
        self.filepath_ply = ''
        self.filepath_project = ''

        eel.init(path_base + 'views', allowed_extensions=['.html'], js_result_timeout=1e9)

        self.root = tkinter.Tk()
        self.screen_width, self.screen_height = self.root.winfo_screenwidth(), self.root.winfo_screenheight()
        self.screen_height -= 50 # taskbar

        self.root.overrideredirect(True)
        self.root.wm_attributes("-topmost", "true")
        self.root.withdraw()

        self.state_adc = StateADC.NotSet

        # root.deiconify()
        # root.overrideredirect(False)
        # root.destroy()

    @staticmethod
    def instance():
        if ControllerChromium.__instance is None:
            ControllerChromium.__instance = ControllerChromium()
        return ControllerChromium.__instance

    def hide_tk_root(self):
        self.root.wm_attributes("-topmost", "false")
        self.root.overrideredirect(False)
        self.root.deiconify()
        self.root.update()
        self.root.overrideredirect(True)
        self.root.wm_attributes("-topmost", "true")
        self.root.withdraw()
        self.root.update()

    def get_default_directory(self, mode: EnumDirectoryMode):
        path_result = path_home = os.path.expanduser('~')
        path_inital = f'{path_home}'
        if os.path.exists(path_inital):
            path_result = path_inital
        if mode == EnumDirectoryMode.MODE_PROJECT:
            if self.filepath_project != '':
                path_result = Path(self.filepath_project).parent
        else:
            if self.filepath_ply != '':
                path_result = Path(self.filepath_ply).parent
            if self.filepath_txt != '':
                path_result = Path(self.filepath_txt).parent
        return path_result

    def update_state_frontend(self):
        eel.update_state_frontend(self.project.to_json())

    def update_state_backend(self, project_json):
        self.project = Project.from_json(project_json)

    def open_filepath_txt(self):
        self.filepath_txt = tkinter.filedialog.askopenfilename(
            parent=self.root,
            title='Open a txt file',
            initialdir=self.get_default_directory(EnumDirectoryMode.MODE_PLY_TXT),
            filetypes=(("txt files", "*.txt"), ("All files", "*.*")))
        self.hide_tk_root()
        return os.path.basename(self.filepath_txt)

    def open_filepath_ply(self):
        self.filepath_ply = tkinter.filedialog.askopenfilename(
            parent=self.root,
            title='Open a PLY file',
            initialdir=self.get_default_directory(EnumDirectoryMode.MODE_PLY_TXT),
            filetypes=(("PLY files", "*.ply"), ("All files", "*.*")))
        self.hide_tk_root()
        return os.path.basename(self.filepath_ply)

    def save_project(self):
        result = ''
        try:
            if self.filepath_project == '' or not os.path.exists(self.filepath_project):
                result = self.save_as_project()
            if self.filepath_project != '':
                result = self.save_project_internal()
        except Exception as e:
            result = str(e)
        return result

    def save_project_internal(self):
        result = ''
        try:
            if len(self.filepath_project):
                project_json = self.project.to_json()
                project_base64 = base64.b64encode(project_json.encode('ascii'))
                with open(self.filepath_project, 'wb') as fp:
                    fp.write(project_base64)
        except Exception as e:
            result = str(e)
        return result

    def save_as_project(self):
        result = ''
        try:
            self.filepath_project = tkinter.filedialog.asksaveasfilename(
                parent=self.root,
                title='Save flight project files',
                initialdir=self.get_default_directory(EnumDirectoryMode.MODE_PROJECT),
                filetypes=(("All files", "*.*"), ("Flight project files", "*.flight"),))
            if not self.filepath_project.endswith('.flight'):
                self.filepath_project += '.flight'
            self.hide_tk_root()
            result = self.save_project_internal()
        except Exception as e:
            result = str(e)

        logger.info(f"save_project result: {result}")
        return result

    def open_project(self):
        result = ''
        try:
            self.filepath_project = tkinter.filedialog.askopenfilename(
                parent=self.root,
                title='Open flight project files',
                initialdir=self.get_default_directory(EnumDirectoryMode.MODE_PROJECT),
                filetypes=(("Flight project files", "*.flight"),))
            self.hide_tk_root()
            if len(self.filepath_project):
                with open(self.filepath_project, 'rb') as fp:
                    project_base64 =  fp.read()
                project_json = base64.b64decode(project_base64.decode('ascii'))
                self.project = Project.from_json(project_json)
                self.update_state_frontend()

        except Exception as e:
            result = str(e)

        logger.info(f"open_project result: {result}")
        return result

    def convert_x_y_to_original(self, x, y):
        result = np.matmul(np.array(self.project.map_layers.transform_to_original), np.array([x, y, 1]))
        return result[0], result[1]

    def convert_x_y_from_original(self, x, y):
        result = np.matmul(np.array(self.project.map_layers.transform_from_original), np.array([x, y, 1]))
        return result[0], result[1]

    def change_poi(self, poi_name, x, y):
        x, y = self.convert_x_y_to_original(x, y)
        if poi_name == "":
            self.add_log_info(
                'UI',
                f'Set position to {x} , {y}'
            )
        else:
            self.add_log_info(
                'UI',
                f'Selected {poi_name} at position {x} , {y}'
            )

    def set_state_adc(self, state: StateADC):
        self.state_adc = state
        eel.set_state_adc(state.value)

    def save_poi(self, poi_name, x, y):
        poi = PointOfInterest(
            uuid=uuid.uuid4(),
            name=poi_name,
            x=x,
            y=y
        )
        self.add_log_info('UI', f'Save POI: {poi_name} {x}, {y}')
        self.project.map_layers.layer_pois.pois.append(poi)
        eel.update_state_frontend(self.project.to_json())

    def delete_poi(self, poi_name):
        self.add_log_info('UI', f'Delete POI: {poi_name}')
        self.project.map_layers.layer_pois.pois = [it for it in self.project.map_layers.layer_pois.pois if it.name != poi_name]
        eel.update_state_frontend(self.project.to_json())

    def clear_all_poi(self):
        self.add_log_info('UI', 'Delete all POIs')
        self.project.map_layers.layer_pois.pois = []
        eel.update_state_frontend(self.project.to_json())

    def set_drone_position(self, x_orig, y_orig, z_orig, dir_x_orig, dir_y_orig, dir_z_orig, dir_w_orig):
        x, y = self.convert_x_y_from_original(x_orig, y_orig)
        angle = 0
        dir_x = 0
        dir_y = 0

        try:
            dir = np.array([dir_x_orig, dir_y_orig])
            dir_norm = np.linalg.norm(dir)
            dir /= dir_norm
            dir_static = np.array([0, -1])
            vector_dot_product = np.dot(dir_static, dir)
            arccos = np.arccos(vector_dot_product)
            angle = np.degrees(arccos)

            dir_x = dir[0]
            dir_y = dir[1]
        except Exception as e:
            logger.exception(e)

        eel.set_drone_position(
            x, y, dir_x, dir_y, angle,
            x_orig, y_orig
        )

    def add_log(
        self,
        timestamp_unix: int,
        source: str,
        severity: LogSeverity,
        message: str
    ):
        new_record = LogRecord(
            timestamp_unix,
            source,
            severity,
            message
        )
        if len(self.project.logs.logs) == 0 or new_record.message != self.project.logs.logs[-1].message:
            logger.info(f'{source} {message}')
            self.project.logs.logs.append(new_record)
            eel.update_state_logs(self.project.logs.to_json())

    def add_log_info(
        self,
        source: str,
        message: str
    ):
        severity = LogSeverity.INFO
        timestamp_unix = int(time.time())
        self.add_log(
            timestamp_unix,
            source,
            severity,
            message
        )

    def load_ply_and_txt(self):
        logger.info("load_ply_and_txt start")

        result = ''
        try:
            project = self.project
            map = load_map_from_ply_and_txt(self.filepath_ply, self.filepath_txt, use_45deg_lines=False)
            map.rotate_and_scale_data()

            project.map_layers.transform_to_original = map.pix_to_meters.tolist()
            project.map_layers.transform_from_original = map.meters_to_pix.tolist()

            project.map_layers.layer_pointcloud_raw.x = [round(it.x, 2) for it in map.point_cloud_in_pix]
            project.map_layers.layer_pointcloud_raw.y = [round(it.y, 2) for it in map.point_cloud_in_pix]

            project.map_layers.layer_pointcloud_filtered.x = [round(it.x, 2) for it in
                                                              map.filtered_point_cloud_in_pix]
            project.map_layers.layer_pointcloud_filtered.y = [round(it.y, 2) for it in
                                                              map.filtered_point_cloud_in_pix]

            project.map_layers.layer_drone_path.x = [round(it.x, 2) for it in map.path_points]
            project.map_layers.layer_drone_path.y = [round(it.y, 2) for it in map.path_points]

            for idx in range(len(map.line_start_points_in_pix)):
                segment_wall = SegmentWall()

                segment_wall.x.append(round(map.line_start_points_in_pix[idx, 0], 2))
                segment_wall.x.append(round(map.line_end_points_in_pix[idx, 0], 2))

                segment_wall.y.append(round(map.line_start_points_in_pix[idx, 1], 2))
                segment_wall.y.append(round(map.line_end_points_in_pix[idx, 1], 2))

                project.map_layers.layer_walls.walls.append(segment_wall)

            eel.update_state_frontend(project.to_json())
        except Exception as e:
            result = str(e)

        logger.info(f"load_ply_and_txt result: {result}")
        return result