import os
import pathlib
import pygubu
import tkinter as tk

from controllers.ControllerADC import ControllerADC
from controllers.ControllerWindowLoader import ControllerDialogLoader
from controllers.ControllerMap import ControllerMap
from controllers.ControllerDesign import ControllerDesign

from models.CommandProps import CommandProps
from models.ModelToggles import ModelToggles


class ControllerWindowMain:
    def __init__(self):
        self.builder = builder = pygubu.Builder()

        root_path = pathlib.Path(__file__).parent.parent
        builder.add_resource_path(root_path)
        builder.add_from_file(f'{root_path}/views/ViewWindowMain.ui')

        self.window_main_ui = builder.get_object('MainWindowToplevel', None)

        self.master = self.window_main_ui
        self.master.title("Alarm Flight UI")
        # maxsize = self.master.maxsize()
        # self.master.geometry(f'{maxsize[0]}x{maxsize[1]}+0+0')

        map_frame = builder.get_object('PlotFrame', self.window_main_ui)
        select_poi_mb = builder.get_object('POISelectPOI', self.window_main_ui)
        delete_poi_btn = builder.get_object('POIDeleteBtn', self.window_main_ui)
        self.controller_map = ControllerMap(master=map_frame,
                                            select_poi_mb=select_poi_mb,
                                            delete_poi_btn=delete_poi_btn)

        self.model_toggles = ModelToggles()
        # Layers
        self.chkRawPointCloud = tk.BooleanVar(value=self.model_toggles.showRawPointcloud)
        self.chkFilteredPointCloud = tk.BooleanVar(value=self.model_toggles.showFilteredPointcloud)
        self.chkWalls = tk.BooleanVar(value=self.model_toggles.showWalls)
        self.chkFligthPath = tk.BooleanVar(value=self.model_toggles.showFligthPath)

        self.viewLayersRawPointcloudBtn = builder.get_object('ViewLayersRawPointcloudBtn', self.master)
        self.viewLayersRawPointcloudBtn.configure(command=self.on_layer_state_change,
                                                  variable=self.chkRawPointCloud)

        self.viewLayersFilteredPointcloudBtn = builder.get_object('ViewLayersFilteredPointcloudBtn', self.master)
        self.viewLayersFilteredPointcloudBtn.configure(command=self.on_layer_state_change,
                                                       variable=self.chkFilteredPointCloud)

        self.viewLayersWallsBtn = builder.get_object('ViewLayersWallsBtn', self.master)
        self.viewLayersWallsBtn.configure(command=self.on_layer_state_change,
                                          variable=self.chkWalls)

        self.viewLayersFlightPathBtn = builder.get_object('ViewLayersFlightPathBtn', self.master)
        self.viewLayersFlightPathBtn.configure(command=self.on_layer_state_change,
                                               variable=self.chkFligthPath)
        # POI
        self.chkShowAllPOI = tk.BooleanVar(value=self.model_toggles.showAllPOIs)
        self.chkSnapToFlightPath = tk.BooleanVar(value=self.model_toggles.snapToFlightPath)

        self.snapToFlightPathBtn = builder.get_object('POISnapFlightPathBtn', self.master)
        self.snapToFlightPathBtn.configure(command=self.on_layer_state_change,
                                           variable=self.chkSnapToFlightPath)

        self.showAllPOIBtn = builder.get_object('POIShowPOIBtn', self.master)
        self.showAllPOIBtn.configure(command=self.on_layer_state_change,
                                     variable=self.chkShowAllPOI)

        self.controller_map.update_toggles(self.model_toggles)

        self.controller_design = ControllerDesign(master=self.master)

        self.controller_design.init_menu_dropdown(label="File", commands=[
            CommandProps(label="Open PLT and TXT..", command=self.on_open_ply_txt),
            CommandProps(label="Exit", command=self.on_close_application)
        ])

        self.controller_design.init_menu_dropdown(label="View", commands=[
            CommandProps(label="Open project", command=lambda: print("oke"))
        ])

        self.controller_design.init_menu_dropdown(label="Help", commands=[
            CommandProps(label="Help", command=lambda: print("oke"))
        ])

        self.master.protocol("WM_DELETE_WINDOW", self.on_close_application)

        button_fly_to_poi: tk.Button = builder.get_object('POIFlyToPOIBtn', self.window_main_ui)
        button_fly_to_poi.configure(
            command=self.on_button_fly_to_poi
        )

        button_action_land: tk.Button = builder.get_object('SidebarControlsLand', self.window_main_ui)
        button_action_land.configure(
            command=self.on_button_action_land
        )

        button_action_take_off: tk.Button = builder.get_object('SidebarControlsTakeoff', self.window_main_ui)
        button_action_take_off.configure(
            command=self.on_button_action_take_off
        )

        button_action_stop: tk.Button = builder.get_object('SidebarControlsStop', self.window_main_ui)
        button_action_stop.configure(
            command=self.on_button_action_stop
        )

        self.builder.connect_callbacks(self)

    def on_button_action_land(self):
        ControllerADC.instance().land()

    def on_button_action_take_off(self):
        ControllerADC.instance().take_off()

    def on_button_action_stop(self):
        ControllerADC.instance().stop()

    def on_button_fly_to_poi(self):
        # TODO add w and z (waiting for info about w)
        ControllerADC.instance().fly_to_poi(
            x=self.controller_map.tmp_point_original_poi[0],
            y=self.controller_map.tmp_point_original_poi[1],
            z=self.controller_map.tmp_point_original_poi[2],
            w=0,
            name_poi="POI"
        )

    def on_close_application(self):
        self.window_main_ui.quit()

    def on_open_ply_txt(self):
        dialog = ControllerDialogLoader(self.controller_map)
        dialog.show(self.window_main_ui)

    def on_layer_state_change(self):
        self.model_toggles.showRawPointcloud = self.chkRawPointCloud.get()
        self.model_toggles.showFilteredPointcloud = self.chkFilteredPointCloud.get()
        self.model_toggles.showWalls = self.chkWalls.get()
        self.model_toggles.showFligthPath = self.chkFligthPath.get()
        self.model_toggles.showAllPOIs = self.chkShowAllPOI.get()
        self.model_toggles.snapToFlightPath = self.chkSnapToFlightPath.get()
        self.controller_map.update_toggles(self.model_toggles)

    def run(self):
        self.window_main_ui.mainloop()
